from typing import Any, Dict, Generic, Type, TypeVar

from .abstract import AbstractElementSerializer

_T = TypeVar('_T')


class IdentitySerializer(AbstractElementSerializer[_T, _T], Generic[_T]):
    def __init__(self, t: Type[_T]):
        self._type = t

    def serialize(self, element: _T) -> _T:
        if not isinstance(element, self._type):
            raise ValueError(f"illegal value {element} of {type(element)}. {self._type} is expected")
        return element

    def deserialize(self, serialized: _T, typed_id2element: Dict[type, Dict[str, Any]]) -> _T:
        if not isinstance(serialized, self._type):
            raise ValueError
        return serialized

    def field_type(self, element_type: Type[_T]) -> Type[_T]:
        return element_type



from enum import Enum
from typing import Optional, Tuple, Union

from pydantic import BaseModel
from typing_extensions import Literal


class DirectiveType(str, Enum):
    CREATE_CONCEPT = 'create_concept'
    CREATE_ACCOUNT = 'create_account'
    CREATE_PLATFORM = 'create_platform'


class AbstractDirectiveModel(BaseModel):
    directive_type: DirectiveType


class CreateConceptDirectiveModel(AbstractDirectiveModel):
    id: str
    name: str
    concept_type: str
    filters: Tuple[dict, ...]
    notes: Optional[str]
    markers: Optional[Tuple[str, ...]]
    access_level: Optional[str]

    directive_type: Literal[DirectiveType.CREATE_CONCEPT] = DirectiveType.CREATE_CONCEPT

    def __hash__(self):
        return hash((self.directive_type, self.concept_type, self.id))


class CreateAccountDirectiveModel(AbstractDirectiveModel):
    key: str
    platform_key: str
    name: str
    url: str

    directive_type: Literal[DirectiveType.CREATE_ACCOUNT] = DirectiveType.CREATE_ACCOUNT

    def __hash__(self):
        return hash((self.directive_type, self.key))


class CreatePlatformDirectiveModel(AbstractDirectiveModel):
    key: str
    platform_type: str
    name: str
    url: str

    directive_type: Literal[DirectiveType.CREATE_PLATFORM] = DirectiveType.CREATE_PLATFORM

    def __hash__(self):
        return hash((self.directive_type, self.key))


DirectiveModel = Union[CreateAccountDirectiveModel, CreateConceptDirectiveModel, CreatePlatformDirectiveModel]



from typing import Optional, Tuple, Union

from tdm.legacy import MetadataModel


class FactMetadataModel(MetadataModel):
    created_time: Optional[int]
    modified_time: Optional[int]
    fact_confidence: Optional[Tuple[float]]
    value_confidence: Union[float, Tuple[float, ...], None]  # same as Optional[float, Tuple[float, ...]] (pydantic bug workaround)

    class Config:
        extra = 'allow'  # any other extra fields will be kept



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Callable, Dict, Generic, Iterable, Iterator, Set, Type, TypeVar, Union

from tdm.helper import generics_mapping, unfold_union
from .identifiable import EnsureIdentifiable, Identifiable

_T = TypeVar('_T', bound='AbstractDomainType')


@dataclass(frozen=True)
class AbstractDomainType(EnsureIdentifiable):
    name: str

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return set()

    @classmethod
    def name_filter(cls: Type[_T], name: str) -> Callable[[_T], bool]:
        def _filter(t: _T) -> bool:
            return t.name == name

        return _filter


_ST = TypeVar('_ST', bound=AbstractDomainType)
_TT = TypeVar('_TT', bound=AbstractDomainType)


@dataclass(frozen=True)
class AbstractLinkDomainType(AbstractDomainType, Generic[_ST, _TT], metaclass=ABCMeta):
    source: _ST
    target: _TT

    def __post_init__(self):
        types_mapping = generics_mapping(type(self))
        if not isinstance(self.source, unfold_union(types_mapping.get(_ST))):
            raise ValueError(f"Illegal source for {type(self)}. Expected: {types_mapping.get(_ST)}, actual: {type(self.source)}")
        if not isinstance(self.target, unfold_union(types_mapping.get(_TT))):
            raise ValueError(f"Illegal target for {type(self)}. Expected: {types_mapping.get(_TT)}, actual: {type(self.target)}")

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'source', 'target'}


_DomainType = TypeVar('_DomainType', bound=AbstractDomainType)


class AbstractDomain(metaclass=ABCMeta):
    __slots__ = ()

    @property
    @abstractmethod
    def id2type(self) -> Dict[str, AbstractDomainType]:
        pass

    @property
    @abstractmethod
    def types(self) -> Dict[Type[AbstractDomainType], Iterable[AbstractDomainType]]:
        pass

    @abstractmethod
    def get_type(self, id_: str) -> AbstractDomainType:
        pass

    @abstractmethod
    def get_types(
            self, type_: Type[_DomainType] = AbstractDomainType, *,
            filter_: Union[Callable[[_DomainType], bool], Iterable[Callable[[_DomainType], bool]]] = tuple()
    ) -> Iterator[_DomainType]:
        pass

    @abstractmethod
    def related_types(
            self, obj: Union[Identifiable, str], type_: Type[_DomainType] = AbstractDomainType, *,
            filter_: Union[Callable[[_DomainType], bool], Iterable[Callable[[_DomainType], bool]]] = tuple()
    ) -> Iterator[_DomainType]:
        pass



from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel.value import AbstractValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class Date(object):
    year: Optional[int] = None
    month: Optional[int] = None
    day: Optional[int] = None


@dataclass(frozen=True)
class Time(object):
    hour: int
    minute: int
    second: int


@dataclass(frozen=True)
class _DateTimeValue(EnsureConfidenced):
    date: Date
    time: Optional[Time] = None


@generate_model
@dataclass(frozen=True)
class DateTimeValue(AbstractValue, _DateTimeValue):

    @classmethod
    def from_dict(cls, value: dict) -> 'DateTimeValue':
        args = {
            'date': Date(**value['date'])
        }
        if 'time' in value:
            args['time'] = Time(**value['time'])
        return cls(**args)



from enum import Enum
from typing import Dict, Optional, Tuple

from pydantic import BaseModel


class NodeType(Enum):
    HEADER = "header"
    TEXT = "text"
    LIST = "list"
    JSON = "json"
    KEY = "key"
    TABLE = "table"
    TABLE_ROW = "row"
    IMAGE = "image"


class NodeMetadata(BaseModel):
    node_type: NodeType
    original_text: Optional[str]
    text_translations: Dict[str, str] = {}
    language: Optional[str] = None
    hidden: bool = False

    class Config:
        extra = 'allow'  # any other extra fields will be kept


class NodeMarkup(BaseModel):
    class Config:
        extra = 'allow'  # any other extra fields will be kept


class TreeDocumentContentModel(BaseModel):
    id: str
    metadata: NodeMetadata
    text: str
    nodes: Optional[Tuple['TreeDocumentContentModel', ...]]
    markup: NodeMarkup = NodeMarkup()


TreeDocumentContentModel.update_forward_refs()



from dataclasses import dataclass

from tdm.abstract.datamodel.value import AbstractValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class _LinkValue(EnsureConfidenced):
    link: str


@generate_model
@dataclass(frozen=True)
class LinkValue(AbstractValue, _LinkValue):
    pass



__all__ = [
    'AbstractDocumentFactory', 'TalismanDocument',
    'AbstractDomainType', 'AbstractLinkDomainType', 'AbstractDomain',
    'AbstractFact', 'AbstractLinkFact', 'FactStatus',
    'and_filter', 'not_filter', 'or_filter',
    'EnsureIdentifiable', 'Identifiable',
    'AbstractNodeLink',
    'AbstractMarkup', 'FrozenMarkup',
    'AbstractNodeMention',
    'AbstractContentNode', 'AbstractNode', 'BaseNodeMetadata',
    'AbstractValue'
]

from .document import AbstractDocumentFactory, TalismanDocument
from .domain import AbstractDomain, AbstractDomainType, AbstractLinkDomainType
from .fact import AbstractFact, AbstractLinkFact, FactStatus
from .filter import and_filter, not_filter, or_filter
from .identifiable import EnsureIdentifiable, Identifiable
from .link import AbstractNodeLink
from .markup import AbstractMarkup, FrozenMarkup
from .mention import AbstractNodeMention
from .node import AbstractContentNode, AbstractNode, BaseNodeMetadata
from .value import AbstractValue



from typing import Dict, Iterator

from tdm.abstract.datamodel import AbstractFact, FactStatus
from tdm.datamodel.domain import PropertyType
from tdm.datamodel.domain.types import AbstractConceptType, AccountType, PlatformType
from tdm.datamodel.facts import AtomValueFact, ConceptFact, PropertyFact
from tdm.datamodel.values import StringValue
from tdm.v0.json_schema.directive import AbstractDirectiveModel, CreateAccountDirectiveModel, CreatePlatformDirectiveModel


def _convert_directive(
        d: AbstractDirectiveModel, type_: AbstractConceptType, properties: Dict[str, PropertyType]
) -> Iterator[AbstractFact]:
    fact = ConceptFact(FactStatus.AUTO, type_)
    yield fact
    for f in d.__fields_set__:
        if f not in properties:
            raise ValueError
        prop_type = properties[f]
        yield PropertyFact(FactStatus.AUTO, prop_type, fact, AtomValueFact(FactStatus.AUTO, prop_type.target, StringValue(getattr(d, f))))


def convert_platform_directive(
        d: CreatePlatformDirectiveModel, type_: PlatformType, properties: Dict[str, PropertyType]
) -> Iterator[AbstractFact]:
    return _convert_directive(d, type_, properties)


def convert_account_directive(
        d: CreateAccountDirectiveModel, type_: AccountType, properties: Dict[str, PropertyType]
) -> Iterator[AbstractFact]:
    return _convert_directive(d, type_, properties)



from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, Iterable, Iterator, Optional, Set, Tuple, Type, TypeVar, Union

from .fact import AbstractFact
from .identifiable import Identifiable
from .link import AbstractNodeLink
from .node import AbstractNode

_TD = TypeVar('_TD', bound='TalismanDocument')
_Fact = TypeVar('_Fact', bound=AbstractFact)
_NodeLink = TypeVar('_NodeLink', bound=AbstractNodeLink)
_Node = TypeVar('_Node', bound=AbstractNode)

NodeOrId = Union[str, AbstractNode]


class TalismanDocument(metaclass=ABCMeta):
    __slots__ = ()

    @property
    @abstractmethod
    def id(self) -> str:
        pass

    # nodes methods

    @property
    @abstractmethod
    def id2node(self) -> Dict[str, AbstractNode]:
        pass

    @property
    @abstractmethod
    def nodes(self) -> Dict[Type[AbstractNode], Iterable[AbstractNode]]:
        pass

    @abstractmethod
    def get_node(self, id_: str) -> AbstractNode:
        pass

    @abstractmethod
    def get_nodes(
            self, type_: Type[_Node] = AbstractNode, *,
            filter_: Union[Callable[[_Node], bool], Iterable[Callable[[_Node], bool]]] = tuple()
    ) -> Iterator[_Node]:
        pass

    @abstractmethod
    def related_nodes(
            self, obj: Union[Identifiable, str], type_: Type[_Node] = AbstractNode, *,
            filter_: Union[Callable[[_Node], bool], Iterable[Callable[[_Node], bool]]] = tuple()
    ) -> Iterator[_Node]:
        pass

    @abstractmethod
    def with_nodes(self: _TD, nodes: Iterable[AbstractNode]) -> _TD:
        pass

    @abstractmethod
    def without_nodes(self: _TD, nodes: Iterable[NodeOrId], *, cascade: bool = True) -> _TD:
        pass

    # structure methods

    @property
    @abstractmethod
    def roots(self) -> Set[AbstractNode]:
        pass

    @property
    @abstractmethod
    def main_root(self) -> Optional[AbstractNode]:
        pass

    @abstractmethod
    def with_main_root(self: _TD, node: Optional[NodeOrId], *, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def parent(self, node: NodeOrId) -> Optional[AbstractNode]:
        pass

    @abstractmethod
    def child_nodes(self, node: Union[str, AbstractNode]) -> Tuple[AbstractNode, ...]:
        pass

    @abstractmethod
    def with_structure(self: _TD, structure: Dict[NodeOrId, Iterable[NodeOrId]], *, force: bool = False, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def with_node_parent(self: _TD, child: NodeOrId, parent: NodeOrId, *, force: bool = False, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def with_roots(self: _TD, nodes: Iterable[NodeOrId]) -> _TD:
        pass

    @abstractmethod
    def without_structure(self: _TD, structure: Dict[NodeOrId, Iterable[NodeOrId]]) -> _TD:
        pass

    # semantic link methods

    @property
    @abstractmethod
    def id2node_link(self) -> Dict[str, AbstractNodeLink]:
        pass

    @property
    @abstractmethod
    def node_links(self) -> Dict[Type[AbstractNodeLink], Iterable[AbstractNodeLink]]:
        pass

    @abstractmethod
    def get_node_link(self, id_: str) -> AbstractNodeLink:
        pass

    @abstractmethod
    def get_node_links(
            self, type_: Type[_NodeLink] = AbstractNodeLink, *,
            filter_: Union[Callable[[_NodeLink], bool], Iterable[Callable[[_NodeLink], bool]]] = tuple()
    ) -> Iterator[_NodeLink]:
        pass

    @abstractmethod
    def related_node_links(
            self, obj: Union[Identifiable, str], type_: Type[_NodeLink] = AbstractNodeLink, *,
            filter_: Union[Callable[[_NodeLink], bool], Iterable[Callable[[_NodeLink], bool]]] = tuple()
    ) -> Iterator[_NodeLink]:
        pass

    @abstractmethod
    def with_node_links(self: _TD, links: Iterable[AbstractNodeLink], *, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def without_node_links(self: _TD, links: Iterable[Union[str, AbstractNodeLink]], *, cascade: bool = False) -> _TD:
        pass

    # facts methods

    @property
    @abstractmethod
    def id2fact(self) -> Dict[str, AbstractFact]:
        pass

    @property
    @abstractmethod
    def facts(self) -> Dict[Type[AbstractFact], Iterable[AbstractFact]]:
        pass

    @abstractmethod
    def get_fact(self, id_: str) -> AbstractFact:
        pass

    @abstractmethod
    def get_facts(
            self, type_: Type[_Fact] = AbstractFact, *,
            filter_: Union[Callable[[_Fact], bool], Iterable[Callable[[_Fact], bool]]] = tuple()
    ) -> Iterator[_Fact]:
        pass

    @abstractmethod
    def related_facts(
            self, obj: Union[Identifiable, str], type_: Type[_Fact] = AbstractFact, *,
            filter_: Union[Callable[[_Fact], bool], Iterable[Callable[[_Fact], bool]]] = tuple()
    ) -> Iterator[_Fact]:
        pass

    @abstractmethod
    def with_facts(self: _TD, facts: Iterable[AbstractFact], *, update: bool = False) -> _TD:
        pass

    @abstractmethod
    def without_facts(self: _TD, facts: Iterable[Union[str, AbstractFact]], *, cascade: bool = False) -> _TD:
        pass


class AbstractDocumentFactory(metaclass=ABCMeta):
    @abstractmethod
    def create_document(self, *, id_: Optional[str] = None, doc_type: Optional[str] = None) -> TalismanDocument:
        pass



from dataclasses import dataclass, field
from typing import Generic, Set, TypeVar

from tdm.helper import generics_mapping
from .identifiable import Identifiable
from .markup import AbstractMarkup, FrozenMarkup


@dataclass(frozen=True)
class BaseNodeMetadata:
    hidden: bool = False


_NodeMetadata = TypeVar("_NodeMetadata", bound=BaseNodeMetadata)


@dataclass(frozen=True)
class AbstractNode(Identifiable, Generic[_NodeMetadata]):
    metadata: _NodeMetadata = None
    markup: AbstractMarkup = field(default_factory=FrozenMarkup, hash=False)

    def __post_init__(self):
        super().__post_init__()
        if self.metadata is None:
            # hack for runtime metadata generation (if no value passed)
            object.__setattr__(self, 'metadata', self._generate_metadata())
        object.__setattr__(self, 'markup', self._convert_markup(self.markup))

    def _convert_markup(self, markup: AbstractMarkup) -> AbstractMarkup:
        if not isinstance(markup, AbstractMarkup):
            raise ValueError
        return markup

    def _generate_metadata(self) -> _NodeMetadata:
        type_vars = generics_mapping(type(self))
        return type_vars[_NodeMetadata]()

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return set()


_Content = TypeVar('_Content')


@dataclass(frozen=True)
class _AbstractContentNode(Generic[_Content]):
    content: _Content


@dataclass(frozen=True)
class AbstractContentNode(AbstractNode[_NodeMetadata], _AbstractContentNode[_Content], Generic[_NodeMetadata, _Content]):

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'content'}



__all__ = [
    'convert_account_directive', 'convert_platform_directive',
    'convert_concept_fact', 'convert_property_fact', 'convert_relation_fact', 'convert_value_fact',
    'get_metadata_facts',
    'build_structure'
]

from ._directives import convert_account_directive, convert_platform_directive
from ._facts import convert_concept_fact, convert_property_fact, convert_relation_fact, convert_value_fact
from ._metadata import get_metadata_facts
from ._nodes import build_structure



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractDomainType, Identifiable


@dataclass(frozen=True)
class AbstractConceptType(Identifiable, AbstractDomainType):
    pass


@dataclass(frozen=True)
class ConceptType(AbstractConceptType):
    pass


@dataclass(frozen=True)
class DocumentType(AbstractConceptType):
    pass


#  following classes could be removed after v0 support stop

@dataclass(frozen=True)
class AccountType(AbstractConceptType):
    pass


@dataclass(frozen=True)
class PlatformType(AbstractConceptType):
    pass



from dataclasses import dataclass, field
from typing import ClassVar, Optional

from tdm.abstract.datamodel import AbstractContentNode, BaseNodeMetadata
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class TextNodeMetadata(BaseNodeMetadata):
    language: Optional[str] = None
    header: bool = False
    header_level: Optional[int] = None

    UNKNOWN_LANG: ClassVar[str] = field(default='unknown', init=False, hash=False, repr=False, compare=False)

    def __post_init__(self):
        if self.header_level is not None:
            if self.header_level < 0:
                raise ValueError(f"header_level could not be negative (actual: {self.header_level})")
            if not self.header:
                raise ValueError("header_level could be specified only with header flag")


@generate_model(label='text')
@dataclass(frozen=True)
class TextNode(AbstractContentNode[TextNodeMetadata, str]):
    pass


@generate_model(label='key')
@dataclass(frozen=True)
class KeyNode(AbstractContentNode[TextNodeMetadata, str]):
    pass



from abc import ABCMeta
from dataclasses import dataclass, replace
from typing import Callable, Sequence, Set, Tuple, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, FactStatus, Identifiable
from tdm.abstract.datamodel.value import AbstractValue
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import AtomValueType, CompositeValueType


@dataclass(frozen=True)
class _AtomValueFact(AbstractFact, metaclass=ABCMeta):
    type_id: Union[str, AtomValueType]
    # tuple is in first place due to strange pydantic bug ((AbstractValue,) is treated as AbstractValue)
    value: Union[Tuple[AbstractValue, ...], AbstractValue] = tuple()

    def __post_init__(self):
        if not isinstance(self.type_id, str) and not isinstance(self.type_id, AtomValueType):
            raise ValueError(f"Illegal type id {self.type_id}, {AtomValueType} is expected")
        value_type = self.type_id.value_type if isinstance(self.type_id, AtomValueType) else AbstractValue

        if isinstance(self.value, Sequence) and not isinstance(self.value, value_type):  # value could implement Sequence?
            if any(not isinstance(v, value_type) for v in self.value):
                raise ValueError(f"Atom value fact {self} value should be {value_type} or tuple of {value_type}")
            object.__setattr__(self, 'value', tuple(self.value))
        elif not isinstance(self.value, value_type):
            raise ValueError(f"Atom value fact {self} value should be value or tuple of value")

        if self.status is FactStatus.NEW and isinstance(self.value, AbstractValue):
            object.__setattr__(self, 'value', (self.value,))
        elif self.status is FactStatus.APPROVED and isinstance(self.value, tuple):
            if len(self.value) != 1 or not isinstance(self.value[0], value_type):
                raise ValueError(f"approved fact {self} should have single value")
            object.__setattr__(self, 'value', self.value[0])

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}


@generate_model(label='atom')
@dataclass(frozen=True)
class AtomValueFact(Identifiable, _AtomValueFact):

    def replace_with_domain(self, domain: AbstractDomain) -> 'AtomValueFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, AtomValueType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, (self.type_id if isinstance(self.type_id, str) else self.type_id.id), self.value

    def __eq__(self, other):
        if not isinstance(other, AtomValueFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    @staticmethod
    def empty_value_filter() -> Callable[['AtomValueFact'], bool]:
        return lambda f: isinstance(f.value, tuple) and not f.value

    @staticmethod
    def tuple_value_filter() -> Callable[['AtomValueFact'], bool]:
        return lambda f: isinstance(f.value, tuple)

    @staticmethod
    def single_value_filter() -> Callable[['AtomValueFact'], bool]:
        return lambda f: isinstance(f.value, AbstractValue)


@dataclass(frozen=True)
class _CompositeValueFact(AbstractFact, metaclass=ABCMeta):
    type_id: Union[str, CompositeValueType]

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}


@generate_model(label='composite')
@dataclass(frozen=True)
class CompositeValueFact(Identifiable, _CompositeValueFact):

    def replace_with_domain(self, domain: AbstractDomain) -> 'CompositeValueFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, CompositeValueType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, (self.type_id if isinstance(self.type_id, str) else self.type_id.id)

    def __eq__(self, other):
        if not isinstance(other, CompositeValueFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())



from dataclasses import dataclass
from typing import Union

from tdm.abstract.datamodel import AbstractLinkFact, Identifiable
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain import PropertyType, RelationPropertyType, RelationType, SlotType
from .concept import ConceptFact
from .value import AtomValueFact, CompositeValueFact


@generate_model(label='relation')
@dataclass(frozen=True, eq=False)
class RelationFact(Identifiable, AbstractLinkFact[ConceptFact, ConceptFact, RelationType]):
    pass


ValueFact = Union[AtomValueFact, CompositeValueFact]


@generate_model(label='property')
@dataclass(frozen=True, eq=False)
class PropertyFact(Identifiable, AbstractLinkFact[ConceptFact, ValueFact, PropertyType]):
    pass


@generate_model(label='r_property')
@dataclass(frozen=True, eq=False)
class RelationPropertyFact(Identifiable, AbstractLinkFact[RelationFact, ValueFact, RelationPropertyType]):
    pass


@generate_model(label='slot')
@dataclass(frozen=True, eq=False)
class SlotFact(Identifiable, AbstractLinkFact[CompositeValueFact, ValueFact, SlotType]):
    pass



from abc import abstractmethod
from dataclasses import dataclass, fields, is_dataclass, make_dataclass
from typing import Callable, Dict, Generic, Set, Tuple, Type, TypeVar, Union

from tdm.abstract.datamodel import EnsureIdentifiable
from tdm.helper import generics_mapping, is_subclass, register_in_module

_T = TypeVar('_T', bound=EnsureIdentifiable)


@dataclass(frozen=True)
class AbstractView(Generic[_T]):
    __depends_on__: Set[str]

    @abstractmethod
    def get_object(self, objects: Dict[str, Union[EnsureIdentifiable, 'AbstractView']]) -> _T:
        pass

    @abstractmethod
    def validate_update(self, other: 'AbstractView[_T]') -> bool:
        pass

    @abstractmethod
    def orig_type(self) -> Type[_T]:
        pass

    @abstractmethod
    def get_dependencies(self, obj: _T) -> Set[EnsureIdentifiable]:
        pass


def generate_depends_on(id_fields: Set[str], view_fields: Set[str]):
    def depends_on(view) -> Set[str]:
        result = {view[field_name] for field_name in id_fields}
        for field_name in view_fields:
            result.update(view[field_name].__depends_on__)
        return result

    return depends_on


_CACHE = {}
_ST = TypeVar('_ST')
_TT = TypeVar('_TT')


def generate_dataclass_view(type_: Type[_ST]) -> Tuple[Type[_TT], Callable[[_ST], _TT]]:
    if type_ in _CACHE:
        return _CACHE[type_]
    if not is_dataclass(type_):
        return type_, lambda x: x

    class_fields = []
    id_fields = set()
    restore_fields = set()
    constructor_fields = {}
    constant_fields = type_.constant_fields() if issubclass(type_, EnsureIdentifiable) else set()
    changed = False
    type_vars = generics_mapping(type_)
    for field in fields(type_):
        field_type = type_vars.get(field.type, field.type)
        if is_subclass(field_type, EnsureIdentifiable):  # we can use id instead of view of object
            field_type = str
            id_fields.add(field.name)
            constructor_fields[field.name] = lambda x: x.id
            changed = True
        else:
            orig_type = field_type
            field_type, field_constructor = generate_dataclass_view(field_type)
            if field_type is not orig_type:
                restore_fields.add(field.name)
                changed = True
                constructor_fields[field.name] = object_view
            else:
                constructor_fields[field.name] = lambda x: x
        class_fields.append((field.name, field_type))

    if not changed and not constant_fields and not issubclass(type_, EnsureIdentifiable):
        _CACHE[type_] = type_, lambda x: x
        return _CACHE[type_]

    def get_object(self, objects: Dict[str, EnsureIdentifiable]) -> type_:
        kwargs = dict(self.__dict__)
        kwargs.pop('__depends_on__')
        for name in id_fields:
            kwargs[name] = restore_object(objects[kwargs[name]], objects)
        for name in restore_fields:
            kwargs[name] = restore_object(kwargs[name], objects)
        return type_(**kwargs)

    def validate_update(self, other) -> None:
        old = self.orig_type()
        new = other.orig_type()
        if not (issubclass(old, new) or issubclass(new, old)):
            raise ValueError(f"changing type of element is forbidden (old type: {old}, new type: {new})")
        for field in constant_fields:
            old = getattr(self, field)
            new = getattr(other, field)
            if old != new:
                raise ValueError(f"changing {field} is forbidden (old value: {old}, new value: {new})")

    def get_dependencies(self, obj) -> Set[EnsureIdentifiable]:
        r = {getattr(obj, field_name) for field_name in id_fields}
        for field_name in restore_fields:
            r.update(getattr(self, field_name).get_dependencies(getattr(obj, field_name)))
        return r

    def eq(self, other) -> bool:
        if not isinstance(other, AbstractView):
            return NotImplemented
        self_orig = self.orig_type()
        other_orig = other.orig_type()
        return (issubclass(self_orig, other_orig) or issubclass(other_orig, self_orig)) and self.__dict__ == other.__dict__

    view_type = make_dataclass(
        cls_name=f"{type_.__name__}View",
        fields=class_fields,
        bases=(AbstractView,),
        eq=False,
        namespace={
            'get_object': get_object,
            'validate_update': validate_update,
            'orig_type': lambda _: type_,
            'get_dependencies': get_dependencies,
            '__eq__': eq
        },
        frozen=True
    )

    view_type = register_in_module(view_type)

    __depends_on__ = generate_depends_on(id_fields, restore_fields)

    def constructor(obj: _ST) -> _TT:
        view = {name: fc(getattr(obj, name)) for name, fc in constructor_fields.items()}
        return view_type(__depends_on__=__depends_on__(view), **view)

    _CACHE[type_] = view_type, constructor
    return _CACHE[type_]


def object_view(obj) -> AbstractView:
    _, constructor = generate_dataclass_view(type(obj))
    return constructor(obj)


def restore_object(view, objects: Dict[str, Union[EnsureIdentifiable, AbstractView]]):
    if isinstance(view, AbstractView):
        return view.get_object(objects)
    return view



from abc import ABCMeta
from dataclasses import dataclass, replace
from typing import Callable, Mapping, Optional, Sequence, Set, Tuple, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, FactStatus, Identifiable
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.domain.types import AbstractConceptType
from tdm.helper import freeze_sequence


@dataclass(frozen=True)
class ConceptValue(object):
    concept: str
    confidence: Optional[float] = None

    def __post_init__(self):
        if self.confidence is not None and not 0 < self.confidence <= 1:
            raise ValueError(f"value confidence should be in interval (0; 1], {self.confidence} is given")

    @classmethod
    def build(cls, value: Union[str, 'ConceptValue']) -> 'ConceptValue':
        if isinstance(value, ConceptValue):
            return value
        if isinstance(value, str):
            return cls(concept=value)
        raise ValueError


@dataclass(frozen=True)
class MissedConceptValue(object):
    filters: Tuple[Mapping[str, object], ...]
    concept_name: str

    def __post_init__(self):
        filters = freeze_sequence(self.filters)
        object.__setattr__(self, 'filters', filters)


@dataclass(frozen=True)
class _ConceptFact(AbstractFact, metaclass=ABCMeta):  # not an error as id argument is kw only
    type_id: Union[str, AbstractConceptType]
    value: Union[ConceptValue, MissedConceptValue, Tuple[ConceptValue, ...]] = tuple()

    def __post_init__(self):
        if not isinstance(self.type_id, str) and not isinstance(self.type_id, AbstractConceptType):
            raise ValueError(f"concept fact type id conflict: {self.type_id}")

        if isinstance(self.value, MissedConceptValue):
            if self.status is FactStatus.APPROVED:
                raise ValueError("approved fact can't be with missed value.")
        elif isinstance(self.value, str):
            object.__setattr__(self, 'value', ConceptValue.build(self.value))
        elif isinstance(self.value, Sequence):
            object.__setattr__(self, 'value', tuple(map(ConceptValue.build, self.value)))
        else:
            object.__setattr__(self, 'value', ConceptValue.build(self.value))

        if self.status is FactStatus.NEW and isinstance(self.value, ConceptValue):
            object.__setattr__(self, 'value', (self.value,))  # replace with tuple
        elif self.status is FactStatus.APPROVED and isinstance(self.value, tuple):
            if len(self.value) != 1:
                raise ValueError(f"approved fact {self} should have single value")
            object.__setattr__(self, 'value', self.value[0])

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}


@generate_model(label='concept')
@dataclass(frozen=True)
class ConceptFact(Identifiable, _ConceptFact):

    def replace_with_domain(self, domain: AbstractDomain) -> 'ConceptFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, AbstractConceptType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, (self.type_id if isinstance(self.type_id, str) else self.type_id.id), self.value

    def __eq__(self, other):
        if not isinstance(other, ConceptFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    @staticmethod
    def empty_value_filter() -> Callable[['ConceptFact'], bool]:
        return lambda f: isinstance(f.value, tuple) and not f.value

    @staticmethod
    def tuple_value_filter() -> Callable[['ConceptFact'], bool]:
        return lambda f: isinstance(f.value, tuple)

    @staticmethod
    def single_value_filter() -> Callable[['ConceptFact'], bool]:
        return lambda f: isinstance(f.value, ConceptValue)



__all__ = [
    'TalismanDocument', 'and_filter', 'not_filter', 'or_filter',
    'DefaultDocumentFactory',
    'TalismanDocumentModel'
]

from .abstract.datamodel import TalismanDocument, and_filter, not_filter, or_filter
from .datamodel.document import DefaultDocumentFactory
from .model import TalismanDocumentModel



__all__ = [
    'DateTimeValue',
    'GeoPointValue',
    'LinkValue',
    'DoubleValue', 'IntValue', 'StringValue'
]

from .date import DateTimeValue
from .geo import GeoPointValue
from .link import LinkValue
from .scalar import DoubleValue, IntValue, StringValue



__all__ = [
    'AbstractNodeWrapper',
    'generate_wrapper', 'composite_markup',
    'validate', 'getter', 'modifier'
]

from ._decorator import MethodType, NodeWrapperDecorator
from ._interface import AbstractNodeWrapper

generate_wrapper = NodeWrapperDecorator.generate_wrapper
composite_markup = NodeWrapperDecorator.composite_markup
validate = NodeWrapperDecorator.validate
post_process = NodeWrapperDecorator.post_process

getter = NodeWrapperDecorator.set_method_type(MethodType.getter)
modifier = NodeWrapperDecorator.set_method_type(MethodType.modifier)



from abc import ABCMeta
from dataclasses import dataclass

from .node import AbstractNode


@dataclass(frozen=True)
class AbstractNodeMention(metaclass=ABCMeta):
    node: AbstractNode

    @property
    def node_id(self) -> str:
        return self.node.id



from enum import Enum
from typing import Any, Optional, Tuple, Union

from pydantic import BaseModel
from typing_extensions import Literal

from tdm.abstract.datamodel import FactStatus
from .metadata import FactMetadataModel


class FactType(str, Enum):
    PROPERTY = "property"
    RELATION = "relation"
    CONCEPT = "concept"
    VALUE = "value"


class SpanModel(BaseModel):
    node_id: str
    value: str
    start: int
    end: int


class AbstractFactModel(BaseModel):
    id: str
    fact_type: FactType
    status: FactStatus
    type_id: str
    value: Any = tuple()  # to support older interface
    mention: Optional[Tuple[SpanModel, ...]]
    metadata: Optional[FactMetadataModel]

    def __hash__(self) -> int:
        return hash((self.id, self.fact_type, self.status, self.type_id))


class ConceptFactModel(AbstractFactModel):
    fact_type: Literal[FactType.CONCEPT] = FactType.CONCEPT


class ValueFactModel(AbstractFactModel):
    fact_type: Literal[FactType.VALUE] = FactType.VALUE


class PropertyLinkValueModel(BaseModel):
    property_id: Optional[str]
    from_fact: str
    to_fact: str


class PropertyFactModel(AbstractFactModel):
    fact_type: Literal[FactType.PROPERTY] = FactType.PROPERTY
    value: PropertyLinkValueModel


class RelationFactModel(AbstractFactModel):
    fact_type: Literal[FactType.RELATION] = FactType.RELATION
    value: PropertyLinkValueModel


FactModel = Union[ConceptFactModel, ValueFactModel, PropertyFactModel, RelationFactModel]



import uuid
from typing import Dict, Iterable, Optional, Tuple

from tdm.abstract.datamodel import AbstractDocumentFactory, AbstractDomain, AbstractFact, AbstractNode, \
    AbstractNodeLink, FactStatus, TalismanDocument
from tdm.datamodel.common import TypedIdsContainer
from tdm.datamodel.domain import get_default_domain
from tdm.datamodel.domain.types import DocumentType
from tdm.datamodel.facts import ConceptFact
from tdm.datamodel.facts.concept import ConceptValue
from ._impl import TalismanDocumentImpl
from ._structure import NodesStructure


class TalismanDocumentFactory(AbstractDocumentFactory):

    def __init__(self, domain: Optional[AbstractDomain] = None):
        self._domain = domain if domain is not None else get_default_domain()

    def create_document(self, *, id_: Optional[str] = None, doc_type: Optional[str] = None) -> TalismanDocument:
        doc = TalismanDocumentImpl(
            id2view={},
            dependencies={},
            structure=NodesStructure(),
            containers={
                AbstractNode: TypedIdsContainer(AbstractNode, ()),
                AbstractNodeLink: TypedIdsContainer(AbstractNodeLink, ()),
                AbstractFact: TypedIdsContainer(AbstractFact, ())
            },
            metadata=None,
            id_=id_ or self.generate_id(),
            domain=self._domain
        )
        if doc_type is not None:
            if self._domain is not None:
                doc_type = self._domain.get_type(doc_type)
                if not isinstance(doc_type, DocumentType):
                    raise ValueError
            return doc.with_facts([ConceptFact(FactStatus.APPROVED, doc_type, ConceptValue(doc.id), id=doc.id)])
        return doc

    def construct(
            self,
            content: Iterable[AbstractNode] = (),
            structure: Dict[str, Tuple[str, ...]] = None,
            root: Optional[str] = None,
            node_links: Iterable[AbstractNodeLink] = (),
            facts: Iterable[AbstractFact] = (),
            *, id_: Optional[str] = None
    ) -> TalismanDocument:
        doc: TalismanDocumentImpl = self.create_document(id_=id_)
        doc = doc.with_elements((*content, *node_links, *facts)) \
            .with_structure(structure) \
            .with_main_root(root)
        return doc

    @staticmethod
    def generate_id():
        return str(uuid.uuid4())



from collections import defaultdict
from typing import Callable, Dict, List, Tuple

from tdm.abstract.datamodel import AbstractNode, AbstractNodeLink, BaseNodeMetadata, FrozenMarkup
from tdm.datamodel.links import TranslationNodeLink
from tdm.datamodel.mentions import NodeMention
from tdm.datamodel.nodes import ImageNode, ImageNodeMetadata, JSONNode, KeyNode, ListNode, ListNodeMetadata, TableCellNode, \
    TableNode, TableRowNode, TableRowNodeMetadata, TextNode, TextNodeMetadata
from tdm.v0.json_schema.content import NodeType, TreeDocumentContentModel


def _convert_text_node(node: TreeDocumentContentModel) -> TextNode:
    return TextNode(
        content=node.text,
        markup=FrozenMarkup.freeze(node.markup.dict()),
        id=node.id,
        metadata=TextNodeMetadata(hidden=node.metadata.hidden, language=node.metadata.language)
    )


def _covert_key_node(node: TreeDocumentContentModel) -> KeyNode:
    return KeyNode(
        content=node.text,
        markup=FrozenMarkup.freeze(node.markup.dict()),
        id=node.id,
        metadata=TextNodeMetadata(hidden=node.metadata.hidden, language=node.metadata.language)
    )


def _convert_list_node(node: TreeDocumentContentModel) -> ListNode:
    return ListNode(id=node.id, metadata=ListNodeMetadata(hidden=node.metadata.hidden))


def _convert_json_node(node: TreeDocumentContentModel) -> JSONNode:
    return JSONNode(id=node.id, metadata=BaseNodeMetadata(hidden=node.metadata.hidden))


def _convert_table_node(node: TreeDocumentContentModel) -> TableNode:
    return TableNode(id=node.id, metadata=BaseNodeMetadata(hidden=node.metadata.hidden))


def _convert_table_row_node(node: TreeDocumentContentModel) -> TableRowNode:
    return TableRowNode(id=node.id, metadata=TableRowNodeMetadata(hidden=node.metadata.hidden))


def _convert_image_node(node: TreeDocumentContentModel) -> ImageNode:
    return ImageNode(
        content=node.text,
        markup=FrozenMarkup.freeze(node.markup.dict()),
        id=node.id,
        metadata=ImageNodeMetadata(hidden=node.metadata.hidden)
    )


NODE_CONVERTERS: Dict[NodeType, Callable[[TreeDocumentContentModel], AbstractNode]] = {
    NodeType.HEADER: _convert_text_node,
    NodeType.TEXT: _convert_text_node,
    NodeType.LIST: _convert_list_node,
    NodeType.JSON: _convert_json_node,
    NodeType.KEY: _covert_key_node,
    NodeType.TABLE: _convert_table_node,
    NodeType.TABLE_ROW: _convert_table_row_node,
    NodeType.IMAGE: _convert_image_node
}


def build_structure(root: TreeDocumentContentModel) -> Tuple[List[AbstractNode], Dict[str, List[str]], List[AbstractNodeLink]]:
    nodes: list[AbstractNode] = []
    structure: dict[str, list[str]] = defaultdict(list)
    node_links: list[AbstractNodeLink] = []

    to_process = [root]  # old document is single tree, so we can start from root

    for node in to_process:
        # create new node for each document node with same type
        converted = NODE_CONVERTERS[node.metadata.node_type](node)
        nodes.append(converted)

        # create separate node with original text (current node is target, original text is source), type is the same
        if node.metadata.original_text is not None:
            if node.metadata.node_type is not NodeType.KEY:
                raise ValueError
            node_links.append(TranslationNodeLink(
                source=NodeMention(KeyNode(content=node.metadata.original_text)),
                target=NodeMention(converted)
            ))

        # text translations are moved to separate nodes linked by translation link (current node is source, translation is target)
        if node.metadata.text_translations:
            if node.metadata.node_type is not NodeType.TEXT and node.metadata.node_type is not NodeType.HEADER:
                raise ValueError
            node_links.extend(
                TranslationNodeLink(
                    source=NodeMention(converted),
                    target=NodeMention(TextNode(content=text, metadata=TextNodeMetadata(language=lang)))
                ) for lang, text in node.metadata.text_translations.items()
            )

        children = node.nodes
        if children:
            if isinstance(converted, TableRowNode):
                # table row should contain only cell nodes (there are no such nodes in old document)
                for c in children:
                    cell = TableCellNode()
                    nodes.append(cell)
                    structure[converted.id].append(cell.id)
                    structure[cell.id].append(c.id)
            else:
                structure[converted.id].extend(map(lambda n: n.id, children))
            to_process.extend(children)

    return nodes, structure, node_links



from dataclasses import dataclass
from typing import Set, Type

from tdm.abstract.datamodel import AbstractDomainType, AbstractValue, Identifiable


@dataclass(frozen=True)
class _AtomValueType(AbstractDomainType):
    value_type: Type[AbstractValue]

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'value_type'}


@dataclass(frozen=True)
class AtomValueType(Identifiable, _AtomValueType):
    pass



from typing import Any, Dict, Type, Union

from tdm.abstract.datamodel import AbstractDomainType
from .abstract import AbstractElementSerializer


class DomainTypeSerializer(AbstractElementSerializer[Union[str, AbstractDomainType], str]):
    def serialize(self, element: Union[str, AbstractDomainType]) -> str:
        return element if isinstance(element, str) else element.id

    def deserialize(self, serialized: str, typed_id2element: Dict[type, Dict[str, Any]]) -> Union[str, AbstractDomainType]:
        return typed_id2element.get(AbstractDomainType, {}).get(serialized, serialized)

    def field_type(self, element_type: Type[Union[str, AbstractDomainType]]) -> Type[str]:
        return str



from abc import ABCMeta
from dataclasses import dataclass
from typing import Optional

from typing_extensions import Self


@dataclass(frozen=True)
class EnsureConfidenced(object):
    def __post_init__(self):
        if not isinstance(self, AbstractValue):
            raise TypeError(f"{type(self)} should inherit {AbstractValue}. Actual mro is {type(self).mro()}")


@dataclass(frozen=True)
class AbstractValue(EnsureConfidenced, metaclass=ABCMeta):
    confidence: Optional[float] = None

    def __post_init__(self):
        if self.confidence is not None and not 0 < self.confidence <= 1:
            raise ValueError(f"value confidence should be in interval (0; 1], {self.confidence} is given")
        for type_ in type(self).mro():
            if issubclass(type_, AbstractValue):
                continue
            if hasattr(type_, '__post_init__'):
                type_.__post_init__(self)

    @classmethod
    def from_dict(cls, value: dict) -> Self:
        return cls(**value)



from immutabledict import immutabledict
from pydantic import BaseModel

from tdm.helper import freeze_dict


class AbstractMarkupModel(BaseModel):
    def immutabledict(self) -> immutabledict:
        return freeze_dict(self.dict(exclude_none=True, exclude_defaults=True))



from abc import abstractmethod
from typing import Type, TypeVar

from immutabledict import immutabledict

_AbstractMarkup = TypeVar('_AbstractMarkup', bound='AbstractMarkup')


class AbstractMarkup(object):
    @property
    @abstractmethod
    def markup(self) -> immutabledict:
        pass

    @classmethod
    @abstractmethod
    def from_markup(cls: Type[_AbstractMarkup], markup: 'AbstractMarkup') -> _AbstractMarkup:
        pass

    def __hash__(self):
        return hash(self.markup)

    def __eq__(self, other):
        if not isinstance(other, AbstractMarkup):
            return NotImplemented
        return self.markup == other.markup



from abc import abstractmethod
from typing import Generic, Type, TypeVar

from tdm.abstract.datamodel import AbstractNode

_Node = TypeVar('_Node', bound=AbstractNode)
_AbstractNodeWrapper = TypeVar('_AbstractNodeWrapper', bound='AbstractNodeWrapper')


class AbstractNodeWrapper(AbstractNode, Generic[_Node]):
    @classmethod
    @abstractmethod
    def wrap(cls: Type[_AbstractNodeWrapper], node: _Node) -> _AbstractNodeWrapper:
        pass



from dataclasses import dataclass
from typing import Callable, Optional

from tdm.abstract.datamodel import AbstractNodeLink, AbstractNodeMention, Identifiable
from tdm.abstract.json_schema import generate_model


@generate_model(label='same')
@dataclass(frozen=True)
class SameNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    pass


@generate_model(label='translation')
@dataclass(frozen=True)
class TranslationNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    language: Optional[str] = None

    @staticmethod
    def lang_filter(language: str) -> Callable[['TranslationNodeLink'], bool]:
        def _filter(link: TranslationNodeLink) -> bool:
            return link.language == language

        return _filter


@generate_model(label='reference')
@dataclass(frozen=True)
class ReferenceNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    pass



from dataclasses import dataclass

from immutabledict import immutabledict

from tdm.helper import freeze_dict
from .abstract import AbstractMarkup


@dataclass(frozen=True)
class FrozenMarkup(AbstractMarkup):
    _markup: immutabledict = immutabledict()

    @property
    def markup(self) -> immutabledict:
        return self._markup

    @classmethod
    def from_markup(cls, markup: AbstractMarkup) -> 'FrozenMarkup':
        if isinstance(markup, FrozenMarkup):
            return markup
        return cls(markup.markup)

    def __hash__(self):
        return hash(self._markup)

    @classmethod
    def freeze(cls, markup: dict) -> 'FrozenMarkup':
        if not isinstance(markup, dict):
            raise ValueError
        return cls(freeze_dict(markup))



from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel.value import AbstractValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class Coordinates(object):
    latitude: float
    longitude: float


@dataclass(frozen=True)
class _GeoPointValue(EnsureConfidenced):
    point: Optional[Coordinates] = None
    name: Optional[str] = None


@generate_model
@dataclass(frozen=True)
class GeoPointValue(AbstractValue, _GeoPointValue):

    @classmethod
    def from_dict(cls, value: dict) -> 'GeoPointValue':
        args = {}
        if 'point' in value:
            args['point'] = Coordinates(**value['point'])
        if 'name' in value:
            args['name'] = value['name']
        return cls(**args)



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Generic, Type, TypeVar

from typing_extensions import Self

_Element = TypeVar('_Element')
_Serialized = TypeVar('_Serialized')


class AbstractElementModel(Generic[_Element]):
    @abstractmethod
    def deserialize(self, typed_id2element: Dict[type, Dict[str, Any]]) -> _Element:
        pass

    @classmethod
    @abstractmethod
    def serialize(cls, element: _Element) -> Self:
        pass


class AbstractElementSerializer(Generic[_Element, _Serialized]):
    @abstractmethod
    def serialize(self, element: _Element) -> _Serialized:
        pass

    @abstractmethod
    def deserialize(self, serialized: _Serialized, typed_id2element: Dict[type, Dict[str, Any]]) -> _Element:
        pass

    @abstractmethod
    def field_type(self, element_type: Type[_Element]) -> Type[_Serialized]:
        pass


class AbstractModelSerializer(AbstractElementSerializer[_Element, AbstractElementModel[_Element]], Generic[_Element], metaclass=ABCMeta):

    def deserialize(self, serialized: AbstractElementModel[_Element], typed_id2element: Dict[type, Dict[str, Any]]) -> _Element:
        return serialized.deserialize(typed_id2element)



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractNodeMention
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.nodes import ImageNode


@generate_model
@dataclass(frozen=True)
class ImageNodeMention(AbstractNodeMention):
    node: ImageNode
    top: int
    bottom: int
    left: int
    right: int

    def __post_init__(self):
        if not isinstance(self.node, ImageNode):
            raise ValueError(f"Incorrect node type {type(self.node)}. Expected {ImageNode}")
        if self.top < 0 or self.bottom <= self.top or self.left < 0 or self.right <= self.left:
            raise ValueError(f"Incorrect bbox [({self.top}, {self.left}); ({self.bottom}, {self.right})]")
        if self.node.metadata.width is not None and self.right > self.node.metadata.width:
            raise ValueError(f"Bbox spreads out of the image (image width: {self.node.metadata.width})")
        if self.node.metadata.height is not None and self.bottom > self.node.metadata.height:
            raise ValueError(f"Bbox spreads out of the image (image height: {self.node.metadata.height})")



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, replace
from enum import Enum
from functools import total_ordering
from typing import Callable, Generic, Iterable, Optional, Set, Type, TypeVar, Union

from tdm.helper import generics_mapping, unfold_union
from .domain import AbstractDomain, AbstractDomainType, AbstractLinkDomainType
from .identifiable import EnsureIdentifiable


@total_ordering
class FactStatus(str, Enum):
    def __new__(cls, name: str, priority: int):
        obj = str.__new__(cls, name)
        obj._value_ = name
        obj.priority = priority
        return obj

    APPROVED = ("approved", 0)
    DECLINED = ("declined", 1)
    AUTO = ("auto", 2)
    HIDDEN = ("hidden", 3)
    NEW = ("new", 4)

    def __lt__(self, other: 'FactStatus'):
        if not isinstance(other, FactStatus):
            return NotImplemented
        return self.priority < other.priority


_Fact = TypeVar('_Fact', bound='AbstractFact')


@dataclass(frozen=True)
class AbstractFact(EnsureIdentifiable, metaclass=ABCMeta):
    status: FactStatus

    @abstractmethod
    def replace_with_domain(self: _Fact, domain: AbstractDomain) -> _Fact:
        pass

    @staticmethod
    def id_filter(obj: Union['AbstractFact', str]) -> Callable[['AbstractFact'], bool]:
        id_ = obj.id if isinstance(obj, AbstractFact) else obj

        def _filter(fact: AbstractFact) -> bool:
            return fact.id == id_

        return _filter

    @staticmethod
    def status_filter(status: Union[FactStatus, Iterable[FactStatus]]) -> Callable[['AbstractFact'], bool]:
        if isinstance(status, FactStatus):
            def _filter(fact: AbstractFact) -> bool:
                return fact.status is status
        else:
            statuses = frozenset(status)

            def _filter(fact: AbstractFact) -> bool:
                return fact.status in statuses

        return _filter

    @staticmethod
    def type_filter(type_: Type['AbstractFact']) -> Callable[['AbstractFact'], bool]:
        def _filter(fact: AbstractFact) -> bool:
            return isinstance(fact, type_)

        return _filter


_ST = TypeVar('_ST', bound=AbstractFact)
_TT = TypeVar('_TT', bound=AbstractFact)
_LDT = TypeVar('_LDT', bound=AbstractLinkDomainType)

_LinkFact = TypeVar('_LinkFact', bound='AbstractLinkFact')


@dataclass(frozen=True)
class AbstractLinkFact(AbstractFact, Generic[_ST, _TT, _LDT], metaclass=ABCMeta):
    type_id: Union[str, _LDT]
    source: _ST
    target: _TT
    value: Optional[str] = None

    def replace_with_domain(self: _LinkFact, domain: AbstractDomain) -> _LinkFact:
        domain_type = generics_mapping(type(self)).get(_LDT)
        type_id = domain.get_type(self.type_id) if isinstance(self.type_id, str) else self.type_id
        if not isinstance(type_id, domain_type):
            raise ValueError

        source = self.source.replace_with_domain(domain)
        target = self.target.replace_with_domain(domain)

        if self.type_id is type_id and self.source is source and self.target is target:
            return self
        return replace(self, type_id=type_id, source=source, target=target)

    def _as_tuple(self) -> tuple:
        return self.id, (self.type_id if isinstance(self.type_id, str) else self.type_id.id), self.source, self.target, self.value

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    def __post_init__(self):
        mapping = generics_mapping(type(self))
        if not isinstance(self.type_id, str):
            domain_type = mapping.get(_LDT)
            if not issubclass(domain_type, AbstractLinkDomainType):
                raise ValueError
            if not isinstance(self.type_id, domain_type):
                raise ValueError(f"illegal type id for fact {self}. {domain_type} is expected")
            self._check_argument(self.type_id.source, self.source.type_id)
            self._check_argument(self.type_id.target, self.target.type_id)
        if not isinstance(self.source, unfold_union(mapping.get(_ST))):
            raise ValueError(f"illegal source fact: {self.source}, {mapping.get(_ST)} is expected")
        if not isinstance(self.target, unfold_union(mapping.get(_TT))):
            raise ValueError(f"illegal target fact: {self.target}, {mapping.get(_TT)} is expected")

    @staticmethod
    def _check_argument(expected_type: AbstractDomainType, actual_type: Union[str, AbstractDomainType]) -> None:
        if isinstance(actual_type, str) and expected_type.id != actual_type:
            raise ValueError
        elif isinstance(actual_type, AbstractDomainType) and expected_type != actual_type:
            raise ValueError

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'source', 'target', 'type_id'}

    @staticmethod
    def type_id_filter(type_id: Union[str, Iterable[str]]) -> Callable[['AbstractLinkFact'], bool]:
        if isinstance(type_id, str):
            def _filter(fact: AbstractLinkFact) -> bool:
                return fact.type_id == type_id
        else:
            type_ids = frozenset(type_id)

            def _filter(fact: AbstractLinkFact) -> bool:
                return fact.type_id in type_ids

        return _filter

    @staticmethod
    def source_filter(filter_: Callable[[_ST], bool]) -> Callable[['AbstractLinkFact'], bool]:
        def _filter(fact: AbstractLinkFact) -> bool:
            return filter_(fact.source)

        return _filter

    @staticmethod
    def target_filter(filter_: Callable[[_TT], bool]) -> Callable[['AbstractLinkFact'], bool]:
        def _filter(fact: AbstractLinkFact) -> bool:
            return filter_(fact.target)

        return _filter



from dataclasses import dataclass
from typing import Set

from tdm.abstract.datamodel import AbstractLinkDomainType, Identifiable
from ._concept import AbstractConceptType


@dataclass(frozen=True)
class _RelationType(AbstractLinkDomainType[AbstractConceptType, AbstractConceptType]):
    directed: bool = True

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return super().constant_fields().union(('directed',))

    pass


@dataclass(frozen=True)
class RelationType(Identifiable, _RelationType):
    pass



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractLinkDomainType, Identifiable
from ._composite import CompositeValueType
from ._value import AtomValueType


@dataclass(frozen=True)
class SlotType(Identifiable, AbstractLinkDomainType[CompositeValueType, AtomValueType]):
    pass



__all__ = [
    'AbstractElementModel', 'AbstractElementSerializer',
    'CompositeElementSerializer',
    'get_serializer'
]

from .abstract import AbstractElementModel, AbstractElementSerializer
from .composite import CompositeElementSerializer
from .serializers import get_serializer



import enum
import inspect
from abc import ABCMeta
from collections import defaultdict
from dataclasses import dataclass, make_dataclass
from typing import Callable, Type, TypeVar, Union

from immutabledict import immutabledict

from tdm.abstract.datamodel import AbstractMarkup, AbstractNode, FrozenMarkup
from tdm.helper import generics_mapping, register_in_module
from ._delegate import getter_delegate, modifier_delegate, property_delegate
from ._impl import AbstractNodeWrapperImpl
from ._interface import _Node as _NodeType

_Node = TypeVar('_Node', bound=AbstractNode)
_Markup = TypeVar('_Markup', bound=AbstractMarkup)


class MethodType(str, enum.Enum):
    getter = 'getter'
    modifier = 'modifier'


class NodeWrapperDecorator(object):
    _validators = defaultdict(list)
    _post_processors = defaultdict(list)
    _method2type = {}

    @classmethod
    def generate_wrapper(cls, markup: Type[_Markup]):
        def generate_type(cls_: Type[_Node]) -> Type[_Node]:
            node_type = generics_mapping(cls_)[_NodeType]
            validators = defaultdict(list)
            post_processors = defaultdict(list)

            for _, f in inspect.getmembers(cls_):
                if not isinstance(f, Callable) and not isinstance(f, property):
                    continue
                if f in cls._validators:
                    for target in cls._validators[f]:
                        validators[target].append(f)
                if f in cls._post_processors:
                    for target in cls._post_processors[f]:
                        post_processors[target].append(f)

            methods = {
                '_node_type': classmethod(lambda _: node_type),
                '_markup_type': classmethod(lambda _: markup),
                '__hash__': lambda self: node_type.__hash__(self)
            }

            for name in cls_.__abstractmethods__:  # delegate all abstract methods to markup
                try:
                    methods[name] = cls._delegate(name, getattr(cls_, name), validators, post_processors)
                except KeyError:
                    pass

            class _Impl(AbstractNodeWrapperImpl[node_type, markup], cls_, metaclass=ABCMeta):
                pass

            return register_in_module(dataclass(frozen=True, eq=False)(type(f'{cls_.__name__}With{markup.__name__}', (_Impl,), methods)))

        return generate_type

    @classmethod
    def composite_markup(cls, **kwargs: Type[AbstractMarkup]):
        fields = tuple(kwargs.items())

        methods_mapping = {}
        for key, type_ in fields:
            for base in type_.mro():
                if issubclass(base, AbstractMarkup) or not hasattr(base, '__abstractmethods__'):
                    continue  # find base interface
                for name in base.__abstractmethods__:
                    if name in methods_mapping:
                        raise TypeError
                    methods_mapping[name] = key

        def generate_composite_markup(cls_: Type[_Markup]) -> Type[_Markup]:
            def markup(self) -> immutabledict:
                return immutabledict(**{
                    key: getattr(self, key).markup for key, _ in fields
                }, **self.__other__)

            def from_markup(cls__, markup: AbstractMarkup):
                markup = dict(markup.markup)
                values = {key: type_.from_markup(FrozenMarkup(_markup=markup.pop(key, immutabledict()))) for key, type_ in fields}
                return cls__(**values, __other__=immutabledict(markup))

            namespace = {
                'markup': property(markup),
                'from_markup': classmethod(from_markup)
            }

            for name in cls_.__abstractmethods__:
                if name not in methods_mapping:
                    continue
                base = methods_mapping[name]
                namespace[name] = cls._delegate(name, getattr(cls_, name), {}, {}, base=base)

            result = register_in_module(make_dataclass(
                f"Generated{cls_.__name__}", [*fields, ('__other__', immutabledict)],
                bases=(cls_,),
                namespace=namespace,
                frozen=True,
                eq=False
            ))

            if not issubclass(result, cls_):
                raise TypeError
            return result

        return generate_composite_markup

    @classmethod
    def validate(cls, orig: Callable):
        def decorate(f):
            # TODO: check signature
            cls._validators[f].append(orig)
            return f

        return decorate

    @classmethod
    def post_process(cls, orig: Union[Callable, property]):
        def decorate(f):
            # TODO: check signature
            cls._post_processors[f].append(orig)
            return f

        return decorate

    @classmethod
    def _delegate(cls, name, f, validators, post_processors, *, base: str = 'markup'):
        if isinstance(f, property):
            return property_delegate(name, post_processors.get(f, []), base=base)
        mode = cls._method2type[f]
        if mode is MethodType.getter:
            return getter_delegate(name, post_processors.get(f, []), base=base)
        elif mode is MethodType.modifier:
            return modifier_delegate(name, validators.get(f, []), base=base)
        raise ValueError

    @classmethod
    def set_method_type(cls, method: MethodType):
        def wrap(f):
            cls._method2type[f] = method
            return f

        return wrap



from operator import attrgetter
from typing import Dict, Iterator, Optional, Sequence, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, FactStatus, Identifiable, TalismanDocument
from tdm.datamodel.domain import AtomValueType, PropertyType
from tdm.datamodel.facts import AtomValueFact, ConceptFact, MentionFact, PropertyFact, RelationFact, RelationPropertyFact
from tdm.datamodel.facts.concept import ConceptValue
from tdm.datamodel.mentions import TextNodeMention
from tdm.datamodel.nodes import TextNode
from tdm.datamodel.values import StringValue
from tdm.v0.json_schema.fact import AbstractFactModel, ConceptFactModel, PropertyFactModel, RelationFactModel, ValueFactModel


def _fact_text(fact: AbstractFactModel) -> Optional[str]:
    if fact.metadata is not None and hasattr(fact.metadata, 'text'):
        return ' '.join(fact.metadata.text)
    if fact.mention:
        return ' '.join(map(attrgetter('value'), fact.mention))
    return None


def convert_concept_fact(
        fact: ConceptFactModel, doc: TalismanDocument, domain: AbstractDomain, titles: Dict[str, PropertyType]
) -> Iterator[AbstractFact]:
    status = FactStatus(fact.status.value)
    if isinstance(fact.value, str):
        value = ConceptValue(fact.value)
    elif isinstance(fact.value, Sequence):
        value = tuple(ConceptValue(v) for v in fact.value)
    else:
        raise ValueError(f"illegal concept fact value {fact.value}")
    concept_fact = ConceptFact(
        status=status,
        type_id=fact.type_id,
        value=value,
        id=fact.id  # preserve old id
    )
    yield concept_fact
    name = _fact_text(fact)
    if not name:
        return
    name_property = titles[fact.type_id]
    if name_property.target.value_type is not StringValue:
        raise ValueError
    old_fashion_value = ValueFactModel(
        id=Identifiable.generate_id(),
        status=fact.status,
        type_id=name_property.target.id,
        value={'value': name},
        mention=fact.mention,
        metadata=fact.metadata
    )
    value, *other = convert_value_fact(old_fashion_value, doc, domain)
    yield value
    yield from other
    yield PropertyFact(
        status=status,
        type_id=name_property,
        source=concept_fact,
        target=value
    )


def convert_value_fact(fact: ValueFactModel, doc: TalismanDocument, domain: AbstractDomain) -> Iterator[AbstractFact]:
    status = FactStatus(fact.status.value)
    value_type = domain.get_type(fact.type_id)
    if not isinstance(value_type, AtomValueType):
        raise ValueError
    if isinstance(fact.value, dict):
        value = value_type.value_type.from_dict(fact.value)
    elif isinstance(fact.value, Sequence):
        value = tuple(value_type.value_type.from_dict(v) for v in fact.value)
    else:
        raise ValueError
    value = AtomValueFact(
        status=status,
        type_id=value_type,
        value=value,
        id=fact.id
    )
    yield value

    if fact.mention:
        # get only first span of multi-span mention
        node = doc.id2node[fact.mention[0].node_id]
        if not isinstance(node, TextNode):
            return
        mention = TextNodeMention(node, fact.mention[0].start, fact.mention[0].end)
    elif fact.metadata and hasattr(fact.metadata, 'text'):
        # create node
        text = ' '.join(fact.metadata.text)
        mention = TextNodeMention(TextNode(text), 0, len(text))
    else:
        return

    yield MentionFact(
        status=status,
        mention=mention,
        value=value
    )


def convert_relation_fact(fact: RelationFactModel, old2new: Dict[str, AbstractFact]) -> RelationFact:
    status = FactStatus(fact.status.value)
    source = old2new[fact.value.from_fact]
    target = old2new[fact.value.to_fact]
    if not isinstance(source, ConceptFact) or not isinstance(target, ConceptFact):
        raise ValueError
    return RelationFact(
        status=status,
        type_id=fact.type_id,
        source=source,
        target=target,
        id=fact.id
    )


def convert_property_fact(fact: PropertyFactModel, old2new: Dict[str, AbstractFact]) -> Union[PropertyFact, RelationPropertyFact]:
    status = FactStatus(fact.status.value)
    source = old2new[fact.value.from_fact]
    target = old2new[fact.value.to_fact]
    if not isinstance(target, AtomValueFact):
        raise ValueError
    if isinstance(source, ConceptFact):
        return PropertyFact(
            status=status,
            type_id=fact.type_id,
            source=source,
            target=target,
            id=fact.id
        )
    elif isinstance(source, RelationFact):
        return RelationPropertyFact(
            status=status,
            type_id=fact.type_id,
            source=source,
            target=target,
            id=fact.id
        )
    raise ValueError



from dataclasses import dataclass
from typing import Callable, Generic, Set, TypeVar

from .identifiable import EnsureIdentifiable
from .mention import AbstractNodeMention

_ST = TypeVar('_ST', bound=AbstractNodeMention)
_TT = TypeVar('_TT', bound=AbstractNodeMention)


@dataclass(frozen=True)
class AbstractNodeLink(EnsureIdentifiable, Generic[_ST, _TT]):
    source: _ST
    target: _TT

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'source', 'target'}

    @staticmethod
    def source_filter(filter_: Callable[[_ST], bool]) -> Callable[['AbstractNodeLink'], bool]:
        def _filter(fact: AbstractNodeLink) -> bool:
            return filter_(fact.source)

        return _filter

    @staticmethod
    def target_filter(filter_: Callable[[_TT], bool]) -> Callable[['AbstractNodeLink'], bool]:
        def _filter(fact: AbstractNodeLink) -> bool:
            return filter_(fact.target)

        return _filter



from typing import Any, Dict, Iterator, Optional, Sequence

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractDomain, AbstractDomainType, AbstractFact, FactStatus
from tdm.datamodel.domain import AtomValueType, PropertyType, RelationType
from tdm.datamodel.domain.types import DocumentType
from tdm.datamodel.facts import AtomValueFact, ConceptFact, PropertyFact, RelationFact
from tdm.datamodel.facts.concept import ConceptValue


def get_metadata_facts(
        doc: TalismanDocument, metadata: Dict[str, Any], domain: Optional[AbstractDomain], metadata_map: Dict[str, str]
) -> Iterator[AbstractFact]:
    document_fact = doc.get_fact(doc.id)
    if not isinstance(document_fact, ConceptFact) or not isinstance(document_fact.type_id, DocumentType):
        raise ValueError

    if not metadata:
        return
    for key, value in metadata.items():
        if key not in metadata_map:
            continue  # maybe log
        domain_types = tuple(domain.related_types(document_fact.type_id, filter_=AbstractDomainType.name_filter(metadata_map[key])))
        if len(domain_types) != 1:
            raise ValueError
        domain_type = domain_types[0]
        if isinstance(domain_type, PropertyType):
            value_type: AtomValueType = domain_type.target
            if not isinstance(value, str) and isinstance(value, Sequence):
                values = tuple(value_type.value_type(v) for v in value)
            else:
                values = (value_type.value_type(value),)
            for v in values:
                yield PropertyFact(FactStatus.APPROVED, domain_type, document_fact, AtomValueFact(FactStatus.APPROVED, value_type, v))
        elif isinstance(domain_type, RelationType):
            yield RelationFact(
                FactStatus.APPROVED, domain_type, document_fact,
                ConceptFact(FactStatus.APPROVED, domain_type.target, ConceptValue(value))
            )
        else:
            raise ValueError



from typing import Optional

from tdm.datamodel.domain._impl import Domain


class DomainManager(object):
    def __init__(self):
        self._default = None

    def set(self, domain: Domain) -> None:
        self._default = domain

    def get(self) -> Optional[Domain]:
        return self._default
